import typing as t

from airflow.api.common.experimental.trigger_dag import trigger_dag
from airflow.models import DagRun
from airflow.operators.trigger_dagrun import TriggerDagRunOperator
from airflow.utils import timezone
from airflow.utils.operator_helpers import determine_kwargs
from airflow.utils.session import provide_session
from airflow.utils.types import DagRunType

from airflow_multi_dagrun.utils import get_multi_dag_run_xcom_key


class TriggerMultiDagRunOperator(TriggerDagRunOperator):

    def __init__(self, op_args=None, op_kwargs=None, python_callable=None, *args, **kwargs):
        super(TriggerMultiDagRunOperator, self).__init__(*args, **kwargs)
        self.op_args = op_args or []
        self.op_kwargs = op_kwargs or {}
        self.python_callable = python_callable

    @provide_session
    def execute(self, context: t.Dict, session=None):
        context.update(self.op_kwargs)
        self.op_kwargs = determine_kwargs(self.python_callable, self.op_args, context)

        created_dr_ids = []
        for conf in self.python_callable(*self.op_args, **self.op_kwargs):
            if not conf:
                break

            execution_date = timezone.utcnow()

            run_id = conf.get('run_id')
            if not run_id:
                run_id = DagRun.generate_run_id(DagRunType.MANUAL, execution_date)

            dag_run = DagRun.find(dag_id=self.trigger_dag_id, run_id=run_id)
            if not dag_run:
                dag_run = trigger_dag(
                    dag_id=self.trigger_dag_id,
                    run_id=run_id,
                    conf=conf,
                    execution_date=execution_date,
                    replace_microseconds=False,
                )
                self.log.info("Created DagRun %s, %s - %s", dag_run, self.trigger_dag_id, run_id)
            else:
                dag_run = dag_run[0]
                self.log.warning("Fetched existed DagRun %s, %s - %s", dag_run, self.trigger_dag_id, run_id)

            created_dr_ids.append(dag_run.id)

        if created_dr_ids:
            xcom_key = get_multi_dag_run_xcom_key(context['execution_date'])
            context['ti'].xcom_push(xcom_key, created_dr_ids)
            self.log.info("Pushed %s DagRun's ids with key %s", len(created_dr_ids), xcom_key)
        else:
            self.log.info("No DagRuns created")



from itertools import chain

from airflow.models import DagRun
from airflow.sensors.base import BaseSensorOperator
from airflow.utils.session import provide_session
from airflow.utils.state import State

from airflow_multi_dagrun.utils import get_multi_dag_run_xcom_key


class ExternalDagsSensor(BaseSensorOperator):
    """
    Wait until the number of running instances is equal to or less than limit

    :param external_dag_id: The dag_id that contains the task you want to
                            wait for
    :type external_dag_id: string
    :param instances_limit: Max number of running instances of DAG
                            with dag_id `external_dag_id`
    :type instances_limit: int
    """
    ui_color = '#19647e'

    def __init__(self, external_dag_id, instances_limit=0, *args, **kwargs):
        super(ExternalDagsSensor, self).__init__(*args, **kwargs)
        self._waited_states = [State.RUNNING]
        self.external_dag_id = external_dag_id
        self._instances_limit = instances_limit

    @provide_session
    def poke(self, context, session=None):
        self.log.info('Poking for {self.external_dag_id}'.format(**locals()))
        count = session.query(DagRun).filter(
            DagRun.dag_id == self.external_dag_id,
            DagRun.state.in_(self._waited_states),
        ).count()
        return count <= self._instances_limit


class MultiDagRunSensor(BaseSensorOperator):
    """
    Used with TriggerMultiDagRunOperator.
    Waits until all DAGs created by TriggerMultiDagRunOperator(in current execution)
    will be finished.
    """

    def __init__(self, dagrun_finished_states=None, *args, **kwargs):
        super(MultiDagRunSensor, self).__init__(*args, **kwargs)
        if dagrun_finished_states is None:
            dagrun_finished_states = [State.SUCCESS, State.FAILED]
        self._dagrun_finished_states = dagrun_finished_states

    @provide_session
    def poke(self, context, session=None):
        xcom_key = get_multi_dag_run_xcom_key(context['execution_date'])
        dagrun_ids = context['ti'].xcom_pull(task_ids=context["task"].upstream_task_ids,
                                             key=xcom_key)
        dagrun_ids = list(chain.from_iterable(dagrun_ids))

        self.log.info("Loaded dagrun ids: %s, for task: %s", dagrun_ids, xcom_key)
        if not dagrun_ids:
            return True

        finished_count = session.query(DagRun).filter(
            DagRun.id.in_(dagrun_ids),
            DagRun.state.in_(self._dagrun_finished_states),
        ).count()
        total_count = len(dagrun_ids)
        self.log.info('Poking for dagruns, finished %s / %s', finished_count, total_count)
        return finished_count >= total_count



from setuptools import setup

from airflow_multi_dagrun import version

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='airflow_multi_dagrun',
    version=version,
    description='MultiDagRunPlugin for airflow',
    python_requires='>=3.6.0',
    author='Ihor Liubymov',
    author_email='infunt@gmail.com',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/mastak/airflow_multi_dagrun',
    packages=['airflow_multi_dagrun'],
    entry_points={
        'airflow.plugins': [
            'airflow_multi_dagrun = airflow_multi_dagrun:MultiDagRunPlugin'
        ]
    }
)
